# -*- coding: utf-8 -*-
from distutils.version import LooseVersion
from itertools import chain

import numpy as np
from numpy import nan
import pytest

from pandas._libs.algos import Infinity, NegInfinity
from pandas._libs.tslib import iNaT
import pandas.compat as compat
from pandas.compat import product
import pandas.util._test_decorators as td

from pandas import NaT, Series, Timestamp, date_range
from pandas.api.types import CategoricalDtype
from pandas.tests.series.common import TestData
import pandas.util.testing as tm
from pandas.util.testing import assert_series_equal


class TestSeriesRank(TestData):
    s = Series([1, 3, 4, 2, nan, 2, 1, 5, nan, 3])

    results = {
        'average': np.array([1.5, 5.5, 7.0, 3.5, nan,
                             3.5, 1.5, 8.0, nan, 5.5]),
        'min': np.array([1, 5, 7, 3, nan, 3, 1, 8, nan, 5]),
        'max': np.array([2, 6, 7, 4, nan, 4, 2, 8, nan, 6]),
        'first': np.array([1, 5, 7, 3, nan, 4, 2, 8, nan, 6]),
        'dense': np.array([1, 3, 4, 2, nan, 2, 1, 5, nan, 3]),
    }

    def test_rank(self):
        pytest.importorskip('scipy.stats.special')
        rankdata = pytest.importorskip('scipy.stats.rankdata')

        self.ts[::2] = np.nan
        self.ts[:10][::3] = 4.

        ranks = self.ts.rank()
        oranks = self.ts.astype('O').rank()

        assert_series_equal(ranks, oranks)

        mask = np.isnan(self.ts)
        filled = self.ts.fillna(np.inf)

        # rankdata returns a ndarray
        exp = Series(rankdata(filled), index=filled.index, name='ts')
        exp[mask] = np.nan

        tm.assert_series_equal(ranks, exp)

        iseries = Series(np.arange(5).repeat(2))

        iranks = iseries.rank()
        exp = iseries.astype(float).rank()
        assert_series_equal(iranks, exp)
        iseries = Series(np.arange(5)) + 1.0
        exp = iseries / 5.0
        iranks = iseries.rank(pct=True)

        assert_series_equal(iranks, exp)

        iseries = Series(np.repeat(1, 100))
        exp = Series(np.repeat(0.505, 100))
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        iseries[1] = np.nan
        exp = Series(np.repeat(50.0 / 99.0, 100))
        exp[1] = np.nan
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        iseries = Series(np.arange(5)) + 1.0
        iseries[4] = np.nan
        exp = iseries / 4.0
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        iseries = Series(np.repeat(np.nan, 100))
        exp = iseries.copy()
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        iseries = Series(np.arange(5)) + 1
        iseries[4] = np.nan
        exp = iseries / 4.0
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        rng = date_range('1/1/1990', periods=5)
        iseries = Series(np.arange(5), rng) + 1
        iseries.iloc[4] = np.nan
        exp = iseries / 4.0
        iranks = iseries.rank(pct=True)
        assert_series_equal(iranks, exp)

        iseries = Series([1e-50, 1e-100, 1e-20, 1e-2, 1e-20 + 1e-30, 1e-1])
        exp = Series([2, 1, 3, 5, 4, 6.0])
        iranks = iseries.rank()
        assert_series_equal(iranks, exp)

        # GH 5968
        iseries = Series(['3 day', '1 day 10m', '-2 day', NaT],
                         dtype='m8[ns]')
        exp = Series([3, 2, 1, np.nan])
        iranks = iseries.rank()
        assert_series_equal(iranks, exp)

        values = np.array(
            [-50, -1, -1e-20, -1e-25, -1e-50, 0, 1e-40, 1e-20, 1e-10, 2, 40
             ], dtype='float64')
        random_order = np.random.permutation(len(values))
        iseries = Series(values[random_order])
        exp = Series(random_order + 1.0, dtype='float64')
        iranks = iseries.rank()
        assert_series_equal(iranks, exp)

    def test_rank_categorical(self):
        # GH issue #15420 rank incorrectly orders ordered categories

        # Test ascending/descending ranking for ordered categoricals
        exp = Series([1., 2., 3., 4., 5., 6.])
        exp_desc = Series([6., 5., 4., 3., 2., 1.])
        ordered = Series(
            ['first', 'second', 'third', 'fourth', 'fifth', 'sixth']
        ).astype(CategoricalDtype(categories=['first', 'second', 'third',
                                              'fourth', 'fifth', 'sixth'],
                                  ordered=True))
        assert_series_equal(ordered.rank(), exp)
        assert_series_equal(ordered.rank(ascending=False), exp_desc)

        # Unordered categoricals should be ranked as objects
        unordered = Series(['first', 'second', 'third', 'fourth',
                            'fifth', 'sixth']).astype(
            CategoricalDtype(categories=['first', 'second', 'third',
                                         'fourth', 'fifth', 'sixth'],
                             ordered=False))
        exp_unordered = Series([2., 4., 6., 3., 1., 5.])
        res = unordered.rank()
        assert_series_equal(res, exp_unordered)

        unordered1 = Series(
            [1, 2, 3, 4, 5, 6],
        ).astype(CategoricalDtype([1, 2, 3, 4, 5, 6], False))
        exp_unordered1 = Series([1., 2., 3., 4., 5., 6.])
        res1 = unordered1.rank()
        assert_series_equal(res1, exp_unordered1)

        # Test na_option for rank data
        na_ser = Series(
            ['first', 'second', 'third', 'fourth', 'fifth', 'sixth', np.NaN]
        ).astype(CategoricalDtype(['first', 'second', 'third', 'fourth',
                                   'fifth', 'sixth', 'seventh'], True))

        exp_top = Series([2., 3., 4., 5., 6., 7., 1.])
        exp_bot = Series([1., 2., 3., 4., 5., 6., 7.])
        exp_keep = Series([1., 2., 3., 4., 5., 6., np.NaN])

        assert_series_equal(na_ser.rank(na_option='top'), exp_top)
        assert_series_equal(na_ser.rank(na_option='bottom'), exp_bot)
        assert_series_equal(na_ser.rank(na_option='keep'), exp_keep)

        # Test na_option for rank data with ascending False
        exp_top = Series([7., 6., 5., 4., 3., 2., 1.])
        exp_bot = Series([6., 5., 4., 3., 2., 1., 7.])
        exp_keep = Series([6., 5., 4., 3., 2., 1., np.NaN])

        assert_series_equal(
            na_ser.rank(na_option='top', ascending=False),
            exp_top
        )
        assert_series_equal(
            na_ser.rank(na_option='bottom', ascending=False),
            exp_bot
        )
        assert_series_equal(
            na_ser.rank(na_option='keep', ascending=False),
            exp_keep
        )

        # Test invalid values for na_option
        msg = "na_option must be one of 'keep', 'top', or 'bottom'"

        with pytest.raises(ValueError, match=msg):
            na_ser.rank(na_option='bad', ascending=False)

        # invalid type
        with pytest.raises(ValueError, match=msg):
            na_ser.rank(na_option=True, ascending=False)

        # Test with pct=True
        na_ser = Series(['first', 'second', 'third', 'fourth', np.NaN]).astype(
            CategoricalDtype(['first', 'second', 'third', 'fourth'], True))
        exp_top = Series([0.4, 0.6, 0.8, 1., 0.2])
        exp_bot = Series([0.2, 0.4, 0.6, 0.8, 1.])
        exp_keep = Series([0.25, 0.5, 0.75, 1., np.NaN])

        assert_series_equal(na_ser.rank(na_option='top', pct=True), exp_top)
        assert_series_equal(na_ser.rank(na_option='bottom', pct=True), exp_bot)
        assert_series_equal(na_ser.rank(na_option='keep', pct=True), exp_keep)

    def test_rank_signature(self):
        s = Series([0, 1])
        s.rank(method='average')
        msg = r"No axis named average for object type <(class|type) 'type'>"
        with pytest.raises(ValueError, match=msg):
            s.rank('average')

    @pytest.mark.parametrize('contents,dtype', [
        ([-np.inf, -50, -1, -1e-20, -1e-25, -1e-50, 0, 1e-40, 1e-20, 1e-10,
          2, 40, np.inf],
         'float64'),
        ([-np.inf, -50, -1, -1e-20, -1e-25, -1e-45, 0, 1e-40, 1e-20, 1e-10,
          2, 40, np.inf],
         'float32'),
        ([np.iinfo(np.uint8).min, 1, 2, 100, np.iinfo(np.uint8).max],
         'uint8'),
        pytest.param([np.iinfo(np.int64).min, -100, 0, 1, 9999, 100000,
                      1e10, np.iinfo(np.int64).max],
                     'int64',
                     marks=pytest.mark.xfail(
                         reason="iNaT is equivalent to minimum value of dtype"
                                "int64 pending issue GH#16674")),
        ([NegInfinity(), '1', 'A', 'BA', 'Ba', 'C', Infinity()],
         'object')
    ])
    def test_rank_inf(self, contents, dtype):
        dtype_na_map = {
            'float64': np.nan,
            'float32': np.nan,
            'int64': iNaT,
            'object': None
        }
        # Insert nans at random positions if underlying dtype has missing
        # value. Then adjust the expected order by adding nans accordingly
        # This is for testing whether rank calculation is affected
        # when values are interwined with nan values.
        values = np.array(contents, dtype=dtype)
        exp_order = np.array(range(len(values)), dtype='float64') + 1.0
        if dtype in dtype_na_map:
            na_value = dtype_na_map[dtype]
            nan_indices = np.random.choice(range(len(values)), 5)
            values = np.insert(values, nan_indices, na_value)
            exp_order = np.insert(exp_order, nan_indices, np.nan)
        # shuffle the testing array and expected results in the same way
        random_order = np.random.permutation(len(values))
        iseries = Series(values[random_order])
        exp = Series(exp_order[random_order], dtype='float64')
        iranks = iseries.rank()
        assert_series_equal(iranks, exp)

    def test_rank_tie_methods(self):
        s = self.s

        def _check(s, expected, method='average'):
            result = s.rank(method=method)
            tm.assert_series_equal(result, Series(expected))

        dtypes = [None, object]
        disabled = {(object, 'first')}
        results = self.results

        for method, dtype in product(results, dtypes):
            if (dtype, method) in disabled:
                continue
            series = s if dtype is None else s.astype(dtype)
            _check(series, results[method], method=method)

    @td.skip_if_no_scipy
    @pytest.mark.parametrize('ascending', [True, False])
    @pytest.mark.parametrize('method', ['average', 'min', 'max', 'first',
                                        'dense'])
    @pytest.mark.parametrize('na_option', ['top', 'bottom', 'keep'])
    def test_rank_tie_methods_on_infs_nans(self, method, na_option, ascending):
        dtypes = [('object', None, Infinity(), NegInfinity()),
                  ('float64', np.nan, np.inf, -np.inf)]
        chunk = 3
        disabled = {('object', 'first')}

        def _check(s, method, na_option, ascending):
            exp_ranks = {
                'average': ([2, 2, 2], [5, 5, 5], [8, 8, 8]),
                'min': ([1, 1, 1], [4, 4, 4], [7, 7, 7]),
                'max': ([3, 3, 3], [6, 6, 6], [9, 9, 9]),
                'first': ([1, 2, 3], [4, 5, 6], [7, 8, 9]),
                'dense': ([1, 1, 1], [2, 2, 2], [3, 3, 3])
            }
            ranks = exp_ranks[method]
            if na_option == 'top':
                order = [ranks[1], ranks[0], ranks[2]]
            elif na_option == 'bottom':
                order = [ranks[0], ranks[2], ranks[1]]
            else:
                order = [ranks[0], [np.nan] * chunk, ranks[1]]
            expected = order if ascending else order[::-1]
            expected = list(chain.from_iterable(expected))
            result = s.rank(method=method, na_option=na_option,
                            ascending=ascending)
            tm.assert_series_equal(result, Series(expected, dtype='float64'))

        for dtype, na_value, pos_inf, neg_inf in dtypes:
            in_arr = [neg_inf] * chunk + [na_value] * chunk + [pos_inf] * chunk
            iseries = Series(in_arr, dtype=dtype)
            if (dtype, method) in disabled:
                continue
            _check(iseries, method, na_option, ascending)

    def test_rank_desc_mix_nans_infs(self):
        # GH 19538
        # check descending ranking when mix nans and infs
        iseries = Series([1, np.nan, np.inf, -np.inf, 25])
        result = iseries.rank(ascending=False)
        exp = Series([3, np.nan, 1, 4, 2], dtype='float64')
        tm.assert_series_equal(result, exp)

    def test_rank_methods_series(self):
        pytest.importorskip('scipy.stats.special')
        rankdata = pytest.importorskip('scipy.stats.rankdata')
        import scipy

        xs = np.random.randn(9)
        xs = np.concatenate([xs[i:] for i in range(0, 9, 2)])  # add duplicates
        np.random.shuffle(xs)

        index = [chr(ord('a') + i) for i in range(len(xs))]

        for vals in [xs, xs + 1e6, xs * 1e-6]:
            ts = Series(vals, index=index)

            for m in ['average', 'min', 'max', 'first', 'dense']:
                result = ts.rank(method=m)
                sprank = rankdata(vals, m if m != 'first' else 'ordinal')
                expected = Series(sprank, index=index)

                if LooseVersion(scipy.__version__) >= LooseVersion('0.17.0'):
                    expected = expected.astype('float64')
                tm.assert_series_equal(result, expected)

    def test_rank_dense_method(self):
        dtypes = ['O', 'f8', 'i8']
        in_out = [([1], [1]),
                  ([2], [1]),
                  ([0], [1]),
                  ([2, 2], [1, 1]),
                  ([1, 2, 3], [1, 2, 3]),
                  ([4, 2, 1], [3, 2, 1],),
                  ([1, 1, 5, 5, 3], [1, 1, 3, 3, 2]),
                  ([-5, -4, -3, -2, -1], [1, 2, 3, 4, 5])]

        for ser, exp in in_out:
            for dtype in dtypes:
                s = Series(ser).astype(dtype)
                result = s.rank(method='dense')
                expected = Series(exp).astype(result.dtype)
                assert_series_equal(result, expected)

    def test_rank_descending(self):
        dtypes = ['O', 'f8', 'i8']

        for dtype, method in product(dtypes, self.results):
            if 'i' in dtype:
                s = self.s.dropna()
            else:
                s = self.s.astype(dtype)

            res = s.rank(ascending=False)
            expected = (s.max() - s).rank()
            assert_series_equal(res, expected)

            if method == 'first' and dtype == 'O':
                continue

            expected = (s.max() - s).rank(method=method)
            res2 = s.rank(method=method, ascending=False)
            assert_series_equal(res2, expected)

    def test_rank_int(self):
        s = self.s.dropna().astype('i8')

        for method, res in compat.iteritems(self.results):
            result = s.rank(method=method)
            expected = Series(res).dropna()
            expected.index = result.index
            assert_series_equal(result, expected)

    def test_rank_object_bug(self):
        # GH 13445

        # smoke tests
        Series([np.nan] * 32).astype(object).rank(ascending=True)
        Series([np.nan] * 32).astype(object).rank(ascending=False)

    def test_rank_modify_inplace(self):
        # GH 18521
        # Check rank does not mutate series
        s = Series([Timestamp('2017-01-05 10:20:27.569000'), NaT])
        expected = s.copy()

        s.rank()
        result = s
        assert_series_equal(result, expected)


# GH15630, pct should be on 100% basis when method='dense'

@pytest.mark.parametrize('dtype', ['O', 'f8', 'i8'])
@pytest.mark.parametrize('ser, exp', [
    ([1], [1.]),
    ([1, 2], [1. / 2, 2. / 2]),
    ([2, 2], [1., 1.]),
    ([1, 2, 3], [1. / 3, 2. / 3, 3. / 3]),
    ([1, 2, 2], [1. / 2, 2. / 2, 2. / 2]),
    ([4, 2, 1], [3. / 3, 2. / 3, 1. / 3],),
    ([1, 1, 5, 5, 3], [1. / 3, 1. / 3, 3. / 3, 3. / 3, 2. / 3]),
    ([1, 1, 3, 3, 5, 5], [1. / 3, 1. / 3, 2. / 3, 2. / 3, 3. / 3, 3. / 3]),
    ([-5, -4, -3, -2, -1], [1. / 5, 2. / 5, 3. / 5, 4. / 5, 5. / 5])])
def test_rank_dense_pct(dtype, ser, exp):
        s = Series(ser).astype(dtype)
        result = s.rank(method='dense', pct=True)
        expected = Series(exp).astype(result.dtype)
        assert_series_equal(result, expected)


@pytest.mark.parametrize('dtype', ['O', 'f8', 'i8'])
@pytest.mark.parametrize('ser, exp', [
    ([1], [1.]),
    ([1, 2], [1. / 2, 2. / 2]),
    ([2, 2], [1. / 2, 1. / 2]),
    ([1, 2, 3], [1. / 3, 2. / 3, 3. / 3]),
    ([1, 2, 2], [1. / 3, 2. / 3, 2. / 3]),
    ([4, 2, 1], [3. / 3, 2. / 3, 1. / 3],),
    ([1, 1, 5, 5, 3], [1. / 5, 1. / 5, 4. / 5, 4. / 5, 3. / 5]),
    ([1, 1, 3, 3, 5, 5], [1. / 6, 1. / 6, 3. / 6, 3. / 6, 5. / 6, 5. / 6]),
    ([-5, -4, -3, -2, -1], [1. / 5, 2. / 5, 3. / 5, 4. / 5, 5. / 5])])
def test_rank_min_pct(dtype, ser, exp):
        s = Series(ser).astype(dtype)
        result = s.rank(method='min', pct=True)
        expected = Series(exp).astype(result.dtype)
        assert_series_equal(result, expected)


@pytest.mark.parametrize('dtype', ['O', 'f8', 'i8'])
@pytest.mark.parametrize('ser, exp', [
    ([1], [1.]),
    ([1, 2], [1. / 2, 2. / 2]),
    ([2, 2], [1., 1.]),
    ([1, 2, 3], [1. / 3, 2. / 3, 3. / 3]),
    ([1, 2, 2], [1. / 3, 3. / 3, 3. / 3]),
    ([4, 2, 1], [3. / 3, 2. / 3, 1. / 3],),
    ([1, 1, 5, 5, 3], [2. / 5, 2. / 5, 5. / 5, 5. / 5, 3. / 5]),
    ([1, 1, 3, 3, 5, 5], [2. / 6, 2. / 6, 4. / 6, 4. / 6, 6. / 6, 6. / 6]),
    ([-5, -4, -3, -2, -1], [1. / 5, 2. / 5, 3. / 5, 4. / 5, 5. / 5])])
def test_rank_max_pct(dtype, ser, exp):
        s = Series(ser).astype(dtype)
        result = s.rank(method='max', pct=True)
        expected = Series(exp).astype(result.dtype)
        assert_series_equal(result, expected)


@pytest.mark.parametrize('dtype', ['O', 'f8', 'i8'])
@pytest.mark.parametrize('ser, exp', [
    ([1], [1.]),
    ([1, 2], [1. / 2, 2. / 2]),
    ([2, 2], [1.5 / 2, 1.5 / 2]),
    ([1, 2, 3], [1. / 3, 2. / 3, 3. / 3]),
    ([1, 2, 2], [1. / 3, 2.5 / 3, 2.5 / 3]),
    ([4, 2, 1], [3. / 3, 2. / 3, 1. / 3],),
    ([1, 1, 5, 5, 3], [1.5 / 5, 1.5 / 5, 4.5 / 5, 4.5 / 5, 3. / 5]),
    ([1, 1, 3, 3, 5, 5],
     [1.5 / 6, 1.5 / 6, 3.5 / 6, 3.5 / 6, 5.5 / 6, 5.5 / 6]),
    ([-5, -4, -3, -2, -1], [1. / 5, 2. / 5, 3. / 5, 4. / 5, 5. / 5])])
def test_rank_average_pct(dtype, ser, exp):
        s = Series(ser).astype(dtype)
        result = s.rank(method='average', pct=True)
        expected = Series(exp).astype(result.dtype)
        assert_series_equal(result, expected)


@pytest.mark.parametrize('dtype', ['f8', 'i8'])
@pytest.mark.parametrize('ser, exp', [
    ([1], [1.]),
    ([1, 2], [1. / 2, 2. / 2]),
    ([2, 2], [1. / 2, 2. / 2.]),
    ([1, 2, 3], [1. / 3, 2. / 3, 3. / 3]),
    ([1, 2, 2], [1. / 3, 2. / 3, 3. / 3]),
    ([4, 2, 1], [3. / 3, 2. / 3, 1. / 3],),
    ([1, 1, 5, 5, 3], [1. / 5, 2. / 5, 4. / 5, 5. / 5, 3. / 5]),
    ([1, 1, 3, 3, 5, 5], [1. / 6, 2. / 6, 3. / 6, 4. / 6, 5. / 6, 6. / 6]),
    ([-5, -4, -3, -2, -1], [1. / 5, 2. / 5, 3. / 5, 4. / 5, 5. / 5])])
def test_rank_first_pct(dtype, ser, exp):
        s = Series(ser).astype(dtype)
        result = s.rank(method='first', pct=True)
        expected = Series(exp).astype(result.dtype)
        assert_series_equal(result, expected)


@pytest.mark.single
def test_pct_max_many_rows():
        # GH 18271
        s = Series(np.arange(2**24 + 1))
        result = s.rank(pct=True).max()
        assert result == 1



# -*- coding: utf-8 -*-
from __future__ import print_function

import operator
import re
from warnings import catch_warnings, simplefilter

import numpy as np
from numpy.random import randn
import pytest

from pandas import _np_version_under1p13, compat
from pandas.core.api import DataFrame, Panel
from pandas.core.computation import expressions as expr
import pandas.util.testing as tm
from pandas.util.testing import (
    assert_almost_equal, assert_frame_equal, assert_panel_equal,
    assert_series_equal)

from pandas.io.formats.printing import pprint_thing

# pylint: disable-msg=W0612,E1101


_frame = DataFrame(randn(10000, 4), columns=list('ABCD'), dtype='float64')
_frame2 = DataFrame(randn(100, 4), columns=list('ABCD'), dtype='float64')
_mixed = DataFrame({'A': _frame['A'].copy(),
                    'B': _frame['B'].astype('float32'),
                    'C': _frame['C'].astype('int64'),
                    'D': _frame['D'].astype('int32')})
_mixed2 = DataFrame({'A': _frame2['A'].copy(),
                     'B': _frame2['B'].astype('float32'),
                     'C': _frame2['C'].astype('int64'),
                     'D': _frame2['D'].astype('int32')})
_integer = DataFrame(
    np.random.randint(1, 100,
                      size=(10001, 4)),
    columns=list('ABCD'), dtype='int64')
_integer2 = DataFrame(np.random.randint(1, 100, size=(101, 4)),
                      columns=list('ABCD'), dtype='int64')

with catch_warnings(record=True):
    simplefilter("ignore", FutureWarning)
    _frame_panel = Panel(dict(ItemA=_frame.copy(),
                              ItemB=(_frame.copy() + 3),
                              ItemC=_frame.copy(),
                              ItemD=_frame.copy()))
    _frame2_panel = Panel(dict(ItemA=_frame2.copy(),
                               ItemB=(_frame2.copy() + 3),
                               ItemC=_frame2.copy(),
                               ItemD=_frame2.copy()))
    _integer_panel = Panel(dict(ItemA=_integer,
                                ItemB=(_integer + 34).astype('int64')))
    _integer2_panel = Panel(dict(ItemA=_integer2,
                                 ItemB=(_integer2 + 34).astype('int64')))
    _mixed_panel = Panel(dict(ItemA=_mixed, ItemB=(_mixed + 3)))
    _mixed2_panel = Panel(dict(ItemA=_mixed2, ItemB=(_mixed2 + 3)))


@pytest.mark.skipif(not expr._USE_NUMEXPR, reason='not using numexpr')
class TestExpressions(object):

    def setup_method(self, method):

        self.frame = _frame.copy()
        self.frame2 = _frame2.copy()
        self.mixed = _mixed.copy()
        self.mixed2 = _mixed2.copy()
        self.integer = _integer.copy()
        self._MIN_ELEMENTS = expr._MIN_ELEMENTS

    def teardown_method(self, method):
        expr._MIN_ELEMENTS = self._MIN_ELEMENTS

    def run_arithmetic(self, df, other, assert_func, check_dtype=False,
                       test_flex=True):
        expr._MIN_ELEMENTS = 0
        operations = ['add', 'sub', 'mul', 'mod', 'truediv', 'floordiv']
        if not compat.PY3:
            operations.append('div')
        for arith in operations:

            operator_name = arith
            if arith == 'div':
                operator_name = 'truediv'

            if test_flex:
                op = lambda x, y: getattr(df, arith)(y)
                op.__name__ = arith
            else:
                op = getattr(operator, operator_name)
            expr.set_use_numexpr(False)
            expected = op(df, other)
            expr.set_use_numexpr(True)

            result = op(df, other)
            try:
                if check_dtype:
                    if arith == 'truediv':
                        assert expected.dtype.kind == 'f'
                assert_func(expected, result)
            except Exception:
                pprint_thing("Failed test with operator %r" % op.__name__)
                raise

    def test_integer_arithmetic(self):
        self.run_arithmetic(self.integer, self.integer,
                            assert_frame_equal)
        self.run_arithmetic(self.integer.iloc[:, 0],
                            self.integer.iloc[:, 0], assert_series_equal,
                            check_dtype=True)

    def run_binary(self, df, other, assert_func, test_flex=False,
                   numexpr_ops={'gt', 'lt', 'ge', 'le', 'eq', 'ne'}):
        """
        tests solely that the result is the same whether or not numexpr is
        enabled.  Need to test whether the function does the correct thing
        elsewhere.
        """
        expr._MIN_ELEMENTS = 0
        expr.set_test_mode(True)
        operations = ['gt', 'lt', 'ge', 'le', 'eq', 'ne']

        for arith in operations:
            if test_flex:
                op = lambda x, y: getattr(df, arith)(y)
                op.__name__ = arith
            else:
                op = getattr(operator, arith)
            expr.set_use_numexpr(False)
            expected = op(df, other)
            expr.set_use_numexpr(True)
            expr.get_test_result()
            result = op(df, other)
            used_numexpr = expr.get_test_result()
            try:
                if arith in numexpr_ops:
                    assert used_numexpr, "Did not use numexpr as expected."
                else:
                    assert not used_numexpr, "Used numexpr unexpectedly."
                assert_func(expected, result)
            except Exception:
                pprint_thing("Failed test with operation %r" % arith)
                pprint_thing("test_flex was %r" % test_flex)
                raise

    def run_frame(self, df, other, binary_comp=None, run_binary=True,
                  **kwargs):
        self.run_arithmetic(df, other, assert_frame_equal,
                            test_flex=False, **kwargs)
        self.run_arithmetic(df, other, assert_frame_equal, test_flex=True,
                            **kwargs)
        if run_binary:
            if binary_comp is None:
                expr.set_use_numexpr(False)
                binary_comp = other + 1
                expr.set_use_numexpr(True)
            self.run_binary(df, binary_comp, assert_frame_equal,
                            test_flex=False, **kwargs)
            self.run_binary(df, binary_comp, assert_frame_equal,
                            test_flex=True, **kwargs)

    def run_series(self, ser, other, binary_comp=None, **kwargs):
        self.run_arithmetic(ser, other, assert_series_equal,
                            test_flex=False, **kwargs)
        self.run_arithmetic(ser, other, assert_almost_equal,
                            test_flex=True, **kwargs)
        # series doesn't uses vec_compare instead of numexpr...
        # if binary_comp is None:
        #     binary_comp = other + 1
        # self.run_binary(ser, binary_comp, assert_frame_equal,
        # test_flex=False, **kwargs)
        # self.run_binary(ser, binary_comp, assert_frame_equal,
        # test_flex=True, **kwargs)

    def run_panel(self, panel, other, binary_comp=None, run_binary=True,
                  assert_func=assert_panel_equal, **kwargs):
        self.run_arithmetic(panel, other, assert_func, test_flex=False,
                            **kwargs)
        self.run_arithmetic(panel, other, assert_func, test_flex=True,
                            **kwargs)
        if run_binary:
            if binary_comp is None:
                binary_comp = other + 1
            self.run_binary(panel, binary_comp, assert_func,
                            test_flex=False, **kwargs)
            self.run_binary(panel, binary_comp, assert_func,
                            test_flex=True, **kwargs)

    def test_integer_arithmetic_frame(self):
        self.run_frame(self.integer, self.integer)

    def test_integer_arithmetic_series(self):
        self.run_series(self.integer.iloc[:, 0], self.integer.iloc[:, 0])

    @pytest.mark.slow
    @pytest.mark.filterwarnings("ignore:\\nPanel:FutureWarning")
    def test_integer_panel(self):
        self.run_panel(_integer2_panel, np.random.randint(1, 100))

    def test_float_arithemtic_frame(self):
        self.run_frame(self.frame2, self.frame2)

    def test_float_arithmetic_series(self):
        self.run_series(self.frame2.iloc[:, 0], self.frame2.iloc[:, 0])

    @pytest.mark.slow
    @pytest.mark.filterwarnings("ignore:\\nPanel:FutureWarning")
    def test_float_panel(self):
        self.run_panel(_frame2_panel, np.random.randn() + 0.1, binary_comp=0.8)

    def test_mixed_arithmetic_frame(self):
        # TODO: FIGURE OUT HOW TO GET IT TO WORK...
        # can't do arithmetic because comparison methods try to do *entire*
        # frame instead of by-column
        self.run_frame(self.mixed2, self.mixed2, run_binary=False)

    def test_mixed_arithmetic_series(self):
        for col in self.mixed2.columns:
            self.run_series(self.mixed2[col], self.mixed2[col], binary_comp=4)

    @pytest.mark.slow
    @pytest.mark.filterwarnings("ignore:\\nPanel:FutureWarning")
    def test_mixed_panel(self):
        self.run_panel(_mixed2_panel, np.random.randint(1, 100),
                       binary_comp=-2)

    def test_float_arithemtic(self):
        self.run_arithmetic(self.frame, self.frame, assert_frame_equal)
        self.run_arithmetic(self.frame.iloc[:, 0], self.frame.iloc[:, 0],
                            assert_series_equal, check_dtype=True)

    def test_mixed_arithmetic(self):
        self.run_arithmetic(self.mixed, self.mixed, assert_frame_equal)
        for col in self.mixed.columns:
            self.run_arithmetic(self.mixed[col], self.mixed[col],
                                assert_series_equal)

    def test_integer_with_zeros(self):
        self.integer *= np.random.randint(0, 2, size=np.shape(self.integer))
        self.run_arithmetic(self.integer, self.integer,
                            assert_frame_equal)
        self.run_arithmetic(self.integer.iloc[:, 0],
                            self.integer.iloc[:, 0], assert_series_equal)

    def test_invalid(self):

        # no op
        result = expr._can_use_numexpr(operator.add, None, self.frame,
                                       self.frame, 'evaluate')
        assert not result

        # mixed
        result = expr._can_use_numexpr(operator.add, '+', self.mixed,
                                       self.frame, 'evaluate')
        assert not result

        # min elements
        result = expr._can_use_numexpr(operator.add, '+', self.frame2,
                                       self.frame2, 'evaluate')
        assert not result

        # ok, we only check on first part of expression
        result = expr._can_use_numexpr(operator.add, '+', self.frame,
                                       self.frame2, 'evaluate')
        assert result

    def test_binary_ops(self):
        def testit():

            for f, f2 in [(self.frame, self.frame2),
                          (self.mixed, self.mixed2)]:

                for op, op_str in [('add', '+'), ('sub', '-'), ('mul', '*'),
                                   ('div', '/'), ('pow', '**')]:

                    if op == 'pow':
                        continue

                    if op == 'div':
                        op = getattr(operator, 'truediv', None)
                    else:
                        op = getattr(operator, op, None)
                    if op is not None:
                        result = expr._can_use_numexpr(op, op_str, f, f,
                                                       'evaluate')
                        assert result != f._is_mixed_type

                        result = expr.evaluate(op, op_str, f, f,
                                               use_numexpr=True)
                        expected = expr.evaluate(op, op_str, f, f,
                                                 use_numexpr=False)

                        if isinstance(result, DataFrame):
                            tm.assert_frame_equal(result, expected)
                        else:
                            tm.assert_numpy_array_equal(result,
                                                        expected.values)

                        result = expr._can_use_numexpr(op, op_str, f2, f2,
                                                       'evaluate')
                        assert not result

        expr.set_use_numexpr(False)
        testit()
        expr.set_use_numexpr(True)
        expr.set_numexpr_threads(1)
        testit()
        expr.set_numexpr_threads()
        testit()

    def test_boolean_ops(self):
        def testit():
            for f, f2 in [(self.frame, self.frame2),
                          (self.mixed, self.mixed2)]:

                f11 = f
                f12 = f + 1

                f21 = f2
                f22 = f2 + 1

                for op, op_str in [('gt', '>'), ('lt', '<'), ('ge', '>='),
                                   ('le', '<='), ('eq', '=='), ('ne', '!=')]:

                    op = getattr(operator, op)

                    result = expr._can_use_numexpr(op, op_str, f11, f12,
                                                   'evaluate')
                    assert result != f11._is_mixed_type

                    result = expr.evaluate(op, op_str, f11, f12,
                                           use_numexpr=True)
                    expected = expr.evaluate(op, op_str, f11, f12,
                                             use_numexpr=False)
                    if isinstance(result, DataFrame):
                        tm.assert_frame_equal(result, expected)
                    else:
                        tm.assert_numpy_array_equal(result, expected.values)

                    result = expr._can_use_numexpr(op, op_str, f21, f22,
                                                   'evaluate')
                    assert not result

        expr.set_use_numexpr(False)
        testit()
        expr.set_use_numexpr(True)
        expr.set_numexpr_threads(1)
        testit()
        expr.set_numexpr_threads()
        testit()

    def test_where(self):
        def testit():
            for f in [self.frame, self.frame2, self.mixed, self.mixed2]:

                for cond in [True, False]:

                    c = np.empty(f.shape, dtype=np.bool_)
                    c.fill(cond)
                    result = expr.where(c, f.values, f.values + 1)
                    expected = np.where(c, f.values, f.values + 1)
                    tm.assert_numpy_array_equal(result, expected)

        expr.set_use_numexpr(False)
        testit()
        expr.set_use_numexpr(True)
        expr.set_numexpr_threads(1)
        testit()
        expr.set_numexpr_threads()
        testit()

    def test_bool_ops_raise_on_arithmetic(self):
        df = DataFrame({'a': np.random.rand(10) > 0.5,
                        'b': np.random.rand(10) > 0.5})
        names = 'div', 'truediv', 'floordiv', 'pow'
        ops = '/', '/', '//', '**'
        msg = 'operator %r not implemented for bool dtypes'
        for op, name in zip(ops, names):
            if not compat.PY3 or name != 'div':
                f = getattr(operator, name)
                err_msg = re.escape(msg % op)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(df, df)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(df.a, df.b)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(df.a, True)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(False, df.a)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(False, df)

                with pytest.raises(NotImplementedError, match=err_msg):
                    f(df, True)

    def test_bool_ops_warn_on_arithmetic(self):
        n = 10
        df = DataFrame({'a': np.random.rand(n) > 0.5,
                        'b': np.random.rand(n) > 0.5})
        names = 'add', 'mul', 'sub'
        ops = '+', '*', '-'
        subs = {'+': '|', '*': '&', '-': '^'}
        sub_funcs = {'|': 'or_', '&': 'and_', '^': 'xor'}
        for op, name in zip(ops, names):
            f = getattr(operator, name)
            fe = getattr(operator, sub_funcs[subs[op]])

            # >= 1.13.0 these are now TypeErrors
            if op == '-' and not _np_version_under1p13:
                continue

            with tm.use_numexpr(True, min_elements=5):
                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(df, df)
                    e = fe(df, df)
                    tm.assert_frame_equal(r, e)

                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(df.a, df.b)
                    e = fe(df.a, df.b)
                    tm.assert_series_equal(r, e)

                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(df.a, True)
                    e = fe(df.a, True)
                    tm.assert_series_equal(r, e)

                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(False, df.a)
                    e = fe(False, df.a)
                    tm.assert_series_equal(r, e)

                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(False, df)
                    e = fe(False, df)
                    tm.assert_frame_equal(r, e)

                with tm.assert_produces_warning(check_stacklevel=False):
                    r = f(df, True)
                    e = fe(df, True)
                    tm.assert_frame_equal(r, e)

    @pytest.mark.parametrize("test_input,expected", [
        (DataFrame([[0, 1, 2, 'aa'], [0, 1, 2, 'aa']],
                   columns=['a', 'b', 'c', 'dtype']),
         DataFrame([[False, False], [False, False]],
                   columns=['a', 'dtype'])),
        (DataFrame([[0, 3, 2, 'aa'], [0, 4, 2, 'aa'], [0, 1, 1, 'bb']],
                   columns=['a', 'b', 'c', 'dtype']),
         DataFrame([[False, False], [False, False],
                   [False, False]], columns=['a', 'dtype'])),
    ])
    def test_bool_ops_column_name_dtype(self, test_input, expected):
        # GH 22383 - .ne fails if columns containing column name 'dtype'
        result = test_input.loc[:, ['a', 'dtype']].ne(
            test_input.loc[:, ['a', 'dtype']])
        assert_frame_equal(result, expected)



from textwrap import dedent

import pytest

from pandas.util._decorators import deprecate

import pandas.util.testing as tm


def new_func():
    """
    This is the summary. The deprecate directive goes next.

    This is the extended summary. The deprecate directive goes before this.
    """
    return 'new_func called'


def new_func_no_docstring():
    return 'new_func_no_docstring called'


def new_func_wrong_docstring():
    """Summary should be in the next line."""
    return 'new_func_wrong_docstring called'


def new_func_with_deprecation():
    """
    This is the summary. The deprecate directive goes next.

    .. deprecated:: 1.0
        Use new_func instead.

    This is the extended summary. The deprecate directive goes before this.
    """
    pass


def test_deprecate_ok():
    depr_func = deprecate('depr_func', new_func, '1.0',
                          msg='Use new_func instead.')

    with tm.assert_produces_warning(FutureWarning):
        result = depr_func()

    assert result == 'new_func called'
    assert depr_func.__doc__ == dedent(new_func_with_deprecation.__doc__)


def test_deprecate_no_docstring():
    depr_func = deprecate('depr_func', new_func_no_docstring, '1.0',
                          msg='Use new_func instead.')
    with tm.assert_produces_warning(FutureWarning):
        result = depr_func()
    assert result == 'new_func_no_docstring called'


def test_deprecate_wrong_docstring():
    with pytest.raises(AssertionError, match='deprecate needs a correctly '
                                             'formatted docstring'):
        deprecate('depr_func', new_func_wrong_docstring, '1.0',
                  msg='Use new_func instead.')



# -*- coding: utf-8 -*-
"""
Tests for Year, Quarter, and Month-based DateOffset subclasses
"""
from datetime import datetime

import pytest

import pandas as pd
from pandas import Timestamp, compat

from pandas.tseries.offsets import (
    BMonthBegin, BMonthEnd, BQuarterBegin, BQuarterEnd, BYearBegin, BYearEnd,
    MonthBegin, MonthEnd, QuarterBegin, QuarterEnd, YearBegin, YearEnd)

from .common import assert_offset_equal, assert_onOffset
from .test_offsets import Base

# --------------------------------------------------------------------
# Misc


def test_quarterly_dont_normalize():
    date = datetime(2012, 3, 31, 5, 30)

    offsets = (QuarterBegin, QuarterEnd, BQuarterEnd, BQuarterBegin)

    for klass in offsets:
        result = date + klass()
        assert (result.time() == date.time())


@pytest.mark.parametrize('n', [-2, 1])
@pytest.mark.parametrize('cls', [MonthBegin, MonthEnd,
                                 BMonthBegin, BMonthEnd,
                                 QuarterBegin, QuarterEnd,
                                 BQuarterBegin, BQuarterEnd,
                                 YearBegin, YearEnd,
                                 BYearBegin, BYearEnd])
def test_apply_index(cls, n):
    offset = cls(n=n)
    rng = pd.date_range(start='1/1/2000', periods=100000, freq='T')
    ser = pd.Series(rng)

    res = rng + offset
    res_v2 = offset.apply_index(rng)
    assert (res == res_v2).all()
    assert res[0] == rng[0] + offset
    assert res[-1] == rng[-1] + offset
    res2 = ser + offset
    # apply_index is only for indexes, not series, so no res2_v2
    assert res2.iloc[0] == ser.iloc[0] + offset
    assert res2.iloc[-1] == ser.iloc[-1] + offset


@pytest.mark.parametrize('offset', [QuarterBegin(), QuarterEnd(),
                                    BQuarterBegin(), BQuarterEnd()])
def test_on_offset(offset):
    dates = [datetime(2016, m, d)
             for m in [10, 11, 12]
             for d in [1, 2, 3, 28, 29, 30, 31] if not (m == 11 and d == 31)]
    for date in dates:
        res = offset.onOffset(date)
        slow_version = date == (date + offset) - offset
        assert res == slow_version


# --------------------------------------------------------------------
# Months

class TestMonthBegin(Base):
    _offset = MonthBegin

    offset_cases = []
    # NOTE: I'm not entirely happy with the logic here for Begin -ss
    # see thread 'offset conventions' on the ML
    offset_cases.append((MonthBegin(), {
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2008, 2, 1): datetime(2008, 3, 1),
        datetime(2006, 12, 31): datetime(2007, 1, 1),
        datetime(2006, 12, 1): datetime(2007, 1, 1),
        datetime(2007, 1, 31): datetime(2007, 2, 1)}))

    offset_cases.append((MonthBegin(0), {
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2006, 12, 3): datetime(2007, 1, 1),
        datetime(2007, 1, 31): datetime(2007, 2, 1)}))

    offset_cases.append((MonthBegin(2), {
        datetime(2008, 2, 29): datetime(2008, 4, 1),
        datetime(2008, 1, 31): datetime(2008, 3, 1),
        datetime(2006, 12, 31): datetime(2007, 2, 1),
        datetime(2007, 12, 28): datetime(2008, 2, 1),
        datetime(2007, 1, 1): datetime(2007, 3, 1),
        datetime(2006, 11, 1): datetime(2007, 1, 1)}))

    offset_cases.append((MonthBegin(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 1),
        datetime(2008, 5, 31): datetime(2008, 5, 1),
        datetime(2008, 12, 31): datetime(2008, 12, 1),
        datetime(2006, 12, 29): datetime(2006, 12, 1),
        datetime(2006, 1, 2): datetime(2006, 1, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)


class TestMonthEnd(Base):
    _offset = MonthEnd

    def test_day_of_month(self):
        dt = datetime(2007, 1, 1)
        offset = MonthEnd()

        result = dt + offset
        assert result == Timestamp(2007, 1, 31)

        result = result + offset
        assert result == Timestamp(2007, 2, 28)

    def test_normalize(self):
        dt = datetime(2007, 1, 1, 3)

        result = dt + MonthEnd(normalize=True)
        expected = dt.replace(hour=0) + MonthEnd()
        assert result == expected

    offset_cases = []
    offset_cases.append((MonthEnd(), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 2, 29),
        datetime(2006, 12, 29): datetime(2006, 12, 31),
        datetime(2006, 12, 31): datetime(2007, 1, 31),
        datetime(2007, 1, 1): datetime(2007, 1, 31),
        datetime(2006, 12, 1): datetime(2006, 12, 31)}))

    offset_cases.append((MonthEnd(0), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 1, 31),
        datetime(2006, 12, 29): datetime(2006, 12, 31),
        datetime(2006, 12, 31): datetime(2006, 12, 31),
        datetime(2007, 1, 1): datetime(2007, 1, 31)}))

    offset_cases.append((MonthEnd(2), {
        datetime(2008, 1, 1): datetime(2008, 2, 29),
        datetime(2008, 1, 31): datetime(2008, 3, 31),
        datetime(2006, 12, 29): datetime(2007, 1, 31),
        datetime(2006, 12, 31): datetime(2007, 2, 28),
        datetime(2007, 1, 1): datetime(2007, 2, 28),
        datetime(2006, 11, 1): datetime(2006, 12, 31)}))

    offset_cases.append((MonthEnd(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 31),
        datetime(2008, 6, 30): datetime(2008, 5, 31),
        datetime(2008, 12, 31): datetime(2008, 11, 30),
        datetime(2006, 12, 29): datetime(2006, 11, 30),
        datetime(2006, 12, 30): datetime(2006, 11, 30),
        datetime(2007, 1, 1): datetime(2006, 12, 31)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(MonthEnd(), datetime(2007, 12, 31), True),
                       (MonthEnd(), datetime(2008, 1, 1), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestBMonthBegin(Base):
    _offset = BMonthBegin

    def test_offsets_compare_equal(self):
        # root cause of #456
        offset1 = BMonthBegin()
        offset2 = BMonthBegin()
        assert not offset1 != offset2

    offset_cases = []
    offset_cases.append((BMonthBegin(), {
        datetime(2008, 1, 1): datetime(2008, 2, 1),
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2006, 12, 29): datetime(2007, 1, 1),
        datetime(2006, 12, 31): datetime(2007, 1, 1),
        datetime(2006, 9, 1): datetime(2006, 10, 2),
        datetime(2007, 1, 1): datetime(2007, 2, 1),
        datetime(2006, 12, 1): datetime(2007, 1, 1)}))

    offset_cases.append((BMonthBegin(0), {
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2006, 10, 2): datetime(2006, 10, 2),
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2006, 12, 29): datetime(2007, 1, 1),
        datetime(2006, 12, 31): datetime(2007, 1, 1),
        datetime(2006, 9, 15): datetime(2006, 10, 2)}))

    offset_cases.append((BMonthBegin(2), {
        datetime(2008, 1, 1): datetime(2008, 3, 3),
        datetime(2008, 1, 15): datetime(2008, 3, 3),
        datetime(2006, 12, 29): datetime(2007, 2, 1),
        datetime(2006, 12, 31): datetime(2007, 2, 1),
        datetime(2007, 1, 1): datetime(2007, 3, 1),
        datetime(2006, 11, 1): datetime(2007, 1, 1)}))

    offset_cases.append((BMonthBegin(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 1),
        datetime(2008, 6, 30): datetime(2008, 6, 2),
        datetime(2008, 6, 1): datetime(2008, 5, 1),
        datetime(2008, 3, 10): datetime(2008, 3, 3),
        datetime(2008, 12, 31): datetime(2008, 12, 1),
        datetime(2006, 12, 29): datetime(2006, 12, 1),
        datetime(2006, 12, 30): datetime(2006, 12, 1),
        datetime(2007, 1, 1): datetime(2006, 12, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(BMonthBegin(), datetime(2007, 12, 31), False),
                       (BMonthBegin(), datetime(2008, 1, 1), True),
                       (BMonthBegin(), datetime(2001, 4, 2), True),
                       (BMonthBegin(), datetime(2008, 3, 3), True)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestBMonthEnd(Base):
    _offset = BMonthEnd

    def test_normalize(self):
        dt = datetime(2007, 1, 1, 3)

        result = dt + BMonthEnd(normalize=True)
        expected = dt.replace(hour=0) + BMonthEnd()
        assert result == expected

    def test_offsets_compare_equal(self):
        # root cause of #456
        offset1 = BMonthEnd()
        offset2 = BMonthEnd()
        assert not offset1 != offset2

    offset_cases = []
    offset_cases.append((BMonthEnd(), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 2, 29),
        datetime(2006, 12, 29): datetime(2007, 1, 31),
        datetime(2006, 12, 31): datetime(2007, 1, 31),
        datetime(2007, 1, 1): datetime(2007, 1, 31),
        datetime(2006, 12, 1): datetime(2006, 12, 29)}))

    offset_cases.append((BMonthEnd(0), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 1, 31),
        datetime(2006, 12, 29): datetime(2006, 12, 29),
        datetime(2006, 12, 31): datetime(2007, 1, 31),
        datetime(2007, 1, 1): datetime(2007, 1, 31)}))

    offset_cases.append((BMonthEnd(2), {
        datetime(2008, 1, 1): datetime(2008, 2, 29),
        datetime(2008, 1, 31): datetime(2008, 3, 31),
        datetime(2006, 12, 29): datetime(2007, 2, 28),
        datetime(2006, 12, 31): datetime(2007, 2, 28),
        datetime(2007, 1, 1): datetime(2007, 2, 28),
        datetime(2006, 11, 1): datetime(2006, 12, 29)}))

    offset_cases.append((BMonthEnd(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 29),
        datetime(2008, 6, 30): datetime(2008, 5, 30),
        datetime(2008, 12, 31): datetime(2008, 11, 28),
        datetime(2006, 12, 29): datetime(2006, 11, 30),
        datetime(2006, 12, 30): datetime(2006, 12, 29),
        datetime(2007, 1, 1): datetime(2006, 12, 29)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(BMonthEnd(), datetime(2007, 12, 31), True),
                       (BMonthEnd(), datetime(2008, 1, 1), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)

# --------------------------------------------------------------------
# Quarters


class TestQuarterBegin(Base):

    def test_repr(self):
        expected = "<QuarterBegin: startingMonth=3>"
        assert repr(QuarterBegin()) == expected
        expected = "<QuarterBegin: startingMonth=3>"
        assert repr(QuarterBegin(startingMonth=3)) == expected
        expected = "<QuarterBegin: startingMonth=1>"
        assert repr(QuarterBegin(startingMonth=1)) == expected

    def test_isAnchored(self):
        assert QuarterBegin(startingMonth=1).isAnchored()
        assert QuarterBegin().isAnchored()
        assert not QuarterBegin(2, startingMonth=1).isAnchored()

    def test_offset_corner_case(self):
        # corner
        offset = QuarterBegin(n=-1, startingMonth=1)
        assert datetime(2010, 2, 1) + offset == datetime(2010, 1, 1)

    offset_cases = []
    offset_cases.append((QuarterBegin(startingMonth=1), {
        datetime(2007, 12, 1): datetime(2008, 1, 1),
        datetime(2008, 1, 1): datetime(2008, 4, 1),
        datetime(2008, 2, 15): datetime(2008, 4, 1),
        datetime(2008, 2, 29): datetime(2008, 4, 1),
        datetime(2008, 3, 15): datetime(2008, 4, 1),
        datetime(2008, 3, 31): datetime(2008, 4, 1),
        datetime(2008, 4, 15): datetime(2008, 7, 1),
        datetime(2008, 4, 1): datetime(2008, 7, 1)}))

    offset_cases.append((QuarterBegin(startingMonth=2), {
        datetime(2008, 1, 1): datetime(2008, 2, 1),
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2008, 1, 15): datetime(2008, 2, 1),
        datetime(2008, 2, 29): datetime(2008, 5, 1),
        datetime(2008, 3, 15): datetime(2008, 5, 1),
        datetime(2008, 3, 31): datetime(2008, 5, 1),
        datetime(2008, 4, 15): datetime(2008, 5, 1),
        datetime(2008, 4, 30): datetime(2008, 5, 1)}))

    offset_cases.append((QuarterBegin(startingMonth=1, n=0), {
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2008, 12, 1): datetime(2009, 1, 1),
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2008, 2, 15): datetime(2008, 4, 1),
        datetime(2008, 2, 29): datetime(2008, 4, 1),
        datetime(2008, 3, 15): datetime(2008, 4, 1),
        datetime(2008, 3, 31): datetime(2008, 4, 1),
        datetime(2008, 4, 15): datetime(2008, 7, 1),
        datetime(2008, 4, 30): datetime(2008, 7, 1)}))

    offset_cases.append((QuarterBegin(startingMonth=1, n=-1), {
        datetime(2008, 1, 1): datetime(2007, 10, 1),
        datetime(2008, 1, 31): datetime(2008, 1, 1),
        datetime(2008, 2, 15): datetime(2008, 1, 1),
        datetime(2008, 2, 29): datetime(2008, 1, 1),
        datetime(2008, 3, 15): datetime(2008, 1, 1),
        datetime(2008, 3, 31): datetime(2008, 1, 1),
        datetime(2008, 4, 15): datetime(2008, 4, 1),
        datetime(2008, 4, 30): datetime(2008, 4, 1),
        datetime(2008, 7, 1): datetime(2008, 4, 1)}))

    offset_cases.append((QuarterBegin(startingMonth=1, n=2), {
        datetime(2008, 1, 1): datetime(2008, 7, 1),
        datetime(2008, 2, 15): datetime(2008, 7, 1),
        datetime(2008, 2, 29): datetime(2008, 7, 1),
        datetime(2008, 3, 15): datetime(2008, 7, 1),
        datetime(2008, 3, 31): datetime(2008, 7, 1),
        datetime(2008, 4, 15): datetime(2008, 10, 1),
        datetime(2008, 4, 1): datetime(2008, 10, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)


class TestQuarterEnd(Base):
    _offset = QuarterEnd

    def test_repr(self):
        expected = "<QuarterEnd: startingMonth=3>"
        assert repr(QuarterEnd()) == expected
        expected = "<QuarterEnd: startingMonth=3>"
        assert repr(QuarterEnd(startingMonth=3)) == expected
        expected = "<QuarterEnd: startingMonth=1>"
        assert repr(QuarterEnd(startingMonth=1)) == expected

    def test_isAnchored(self):
        assert QuarterEnd(startingMonth=1).isAnchored()
        assert QuarterEnd().isAnchored()
        assert not QuarterEnd(2, startingMonth=1).isAnchored()

    def test_offset_corner_case(self):
        # corner
        offset = QuarterEnd(n=-1, startingMonth=1)
        assert datetime(2010, 2, 1) + offset == datetime(2010, 1, 31)

    offset_cases = []
    offset_cases.append((QuarterEnd(startingMonth=1), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 4, 30),
        datetime(2008, 2, 15): datetime(2008, 4, 30),
        datetime(2008, 2, 29): datetime(2008, 4, 30),
        datetime(2008, 3, 15): datetime(2008, 4, 30),
        datetime(2008, 3, 31): datetime(2008, 4, 30),
        datetime(2008, 4, 15): datetime(2008, 4, 30),
        datetime(2008, 4, 30): datetime(2008, 7, 31)}))

    offset_cases.append((QuarterEnd(startingMonth=2), {
        datetime(2008, 1, 1): datetime(2008, 2, 29),
        datetime(2008, 1, 31): datetime(2008, 2, 29),
        datetime(2008, 2, 15): datetime(2008, 2, 29),
        datetime(2008, 2, 29): datetime(2008, 5, 31),
        datetime(2008, 3, 15): datetime(2008, 5, 31),
        datetime(2008, 3, 31): datetime(2008, 5, 31),
        datetime(2008, 4, 15): datetime(2008, 5, 31),
        datetime(2008, 4, 30): datetime(2008, 5, 31)}))

    offset_cases.append((QuarterEnd(startingMonth=1, n=0), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 1, 31),
        datetime(2008, 2, 15): datetime(2008, 4, 30),
        datetime(2008, 2, 29): datetime(2008, 4, 30),
        datetime(2008, 3, 15): datetime(2008, 4, 30),
        datetime(2008, 3, 31): datetime(2008, 4, 30),
        datetime(2008, 4, 15): datetime(2008, 4, 30),
        datetime(2008, 4, 30): datetime(2008, 4, 30)}))

    offset_cases.append((QuarterEnd(startingMonth=1, n=-1), {
        datetime(2008, 1, 1): datetime(2007, 10, 31),
        datetime(2008, 1, 31): datetime(2007, 10, 31),
        datetime(2008, 2, 15): datetime(2008, 1, 31),
        datetime(2008, 2, 29): datetime(2008, 1, 31),
        datetime(2008, 3, 15): datetime(2008, 1, 31),
        datetime(2008, 3, 31): datetime(2008, 1, 31),
        datetime(2008, 4, 15): datetime(2008, 1, 31),
        datetime(2008, 4, 30): datetime(2008, 1, 31),
        datetime(2008, 7, 1): datetime(2008, 4, 30)}))

    offset_cases.append((QuarterEnd(startingMonth=1, n=2), {
        datetime(2008, 1, 31): datetime(2008, 7, 31),
        datetime(2008, 2, 15): datetime(2008, 7, 31),
        datetime(2008, 2, 29): datetime(2008, 7, 31),
        datetime(2008, 3, 15): datetime(2008, 7, 31),
        datetime(2008, 3, 31): datetime(2008, 7, 31),
        datetime(2008, 4, 15): datetime(2008, 7, 31),
        datetime(2008, 4, 30): datetime(2008, 10, 31)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [
        (QuarterEnd(1, startingMonth=1), datetime(2008, 1, 31), True),
        (QuarterEnd(1, startingMonth=1), datetime(2007, 12, 31), False),
        (QuarterEnd(1, startingMonth=1), datetime(2008, 2, 29), False),
        (QuarterEnd(1, startingMonth=1), datetime(2007, 3, 30), False),
        (QuarterEnd(1, startingMonth=1), datetime(2007, 3, 31), False),
        (QuarterEnd(1, startingMonth=1), datetime(2008, 4, 30), True),
        (QuarterEnd(1, startingMonth=1), datetime(2008, 5, 30), False),
        (QuarterEnd(1, startingMonth=1), datetime(2008, 5, 31), False),
        (QuarterEnd(1, startingMonth=1), datetime(2007, 6, 29), False),
        (QuarterEnd(1, startingMonth=1), datetime(2007, 6, 30), False),
        (QuarterEnd(1, startingMonth=2), datetime(2008, 1, 31), False),
        (QuarterEnd(1, startingMonth=2), datetime(2007, 12, 31), False),
        (QuarterEnd(1, startingMonth=2), datetime(2008, 2, 29), True),
        (QuarterEnd(1, startingMonth=2), datetime(2007, 3, 30), False),
        (QuarterEnd(1, startingMonth=2), datetime(2007, 3, 31), False),
        (QuarterEnd(1, startingMonth=2), datetime(2008, 4, 30), False),
        (QuarterEnd(1, startingMonth=2), datetime(2008, 5, 30), False),
        (QuarterEnd(1, startingMonth=2), datetime(2008, 5, 31), True),
        (QuarterEnd(1, startingMonth=2), datetime(2007, 6, 29), False),
        (QuarterEnd(1, startingMonth=2), datetime(2007, 6, 30), False),
        (QuarterEnd(1, startingMonth=3), datetime(2008, 1, 31), False),
        (QuarterEnd(1, startingMonth=3), datetime(2007, 12, 31), True),
        (QuarterEnd(1, startingMonth=3), datetime(2008, 2, 29), False),
        (QuarterEnd(1, startingMonth=3), datetime(2007, 3, 30), False),
        (QuarterEnd(1, startingMonth=3), datetime(2007, 3, 31), True),
        (QuarterEnd(1, startingMonth=3), datetime(2008, 4, 30), False),
        (QuarterEnd(1, startingMonth=3), datetime(2008, 5, 30), False),
        (QuarterEnd(1, startingMonth=3), datetime(2008, 5, 31), False),
        (QuarterEnd(1, startingMonth=3), datetime(2007, 6, 29), False),
        (QuarterEnd(1, startingMonth=3), datetime(2007, 6, 30), True)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestBQuarterBegin(Base):
    _offset = BQuarterBegin

    def test_repr(self):
        expected = "<BusinessQuarterBegin: startingMonth=3>"
        assert repr(BQuarterBegin()) == expected
        expected = "<BusinessQuarterBegin: startingMonth=3>"
        assert repr(BQuarterBegin(startingMonth=3)) == expected
        expected = "<BusinessQuarterBegin: startingMonth=1>"
        assert repr(BQuarterBegin(startingMonth=1)) == expected

    def test_isAnchored(self):
        assert BQuarterBegin(startingMonth=1).isAnchored()
        assert BQuarterBegin().isAnchored()
        assert not BQuarterBegin(2, startingMonth=1).isAnchored()

    def test_offset_corner_case(self):
        # corner
        offset = BQuarterBegin(n=-1, startingMonth=1)
        assert datetime(2007, 4, 3) + offset == datetime(2007, 4, 2)

    offset_cases = []
    offset_cases.append((BQuarterBegin(startingMonth=1), {
        datetime(2008, 1, 1): datetime(2008, 4, 1),
        datetime(2008, 1, 31): datetime(2008, 4, 1),
        datetime(2008, 2, 15): datetime(2008, 4, 1),
        datetime(2008, 2, 29): datetime(2008, 4, 1),
        datetime(2008, 3, 15): datetime(2008, 4, 1),
        datetime(2008, 3, 31): datetime(2008, 4, 1),
        datetime(2008, 4, 15): datetime(2008, 7, 1),
        datetime(2007, 3, 15): datetime(2007, 4, 2),
        datetime(2007, 2, 28): datetime(2007, 4, 2),
        datetime(2007, 1, 1): datetime(2007, 4, 2),
        datetime(2007, 4, 15): datetime(2007, 7, 2),
        datetime(2007, 7, 1): datetime(2007, 7, 2),
        datetime(2007, 4, 1): datetime(2007, 4, 2),
        datetime(2007, 4, 2): datetime(2007, 7, 2),
        datetime(2008, 4, 30): datetime(2008, 7, 1)}))

    offset_cases.append((BQuarterBegin(startingMonth=2), {
        datetime(2008, 1, 1): datetime(2008, 2, 1),
        datetime(2008, 1, 31): datetime(2008, 2, 1),
        datetime(2008, 1, 15): datetime(2008, 2, 1),
        datetime(2008, 2, 29): datetime(2008, 5, 1),
        datetime(2008, 3, 15): datetime(2008, 5, 1),
        datetime(2008, 3, 31): datetime(2008, 5, 1),
        datetime(2008, 4, 15): datetime(2008, 5, 1),
        datetime(2008, 8, 15): datetime(2008, 11, 3),
        datetime(2008, 9, 15): datetime(2008, 11, 3),
        datetime(2008, 11, 1): datetime(2008, 11, 3),
        datetime(2008, 4, 30): datetime(2008, 5, 1)}))

    offset_cases.append((BQuarterBegin(startingMonth=1, n=0), {
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2007, 12, 31): datetime(2008, 1, 1),
        datetime(2008, 2, 15): datetime(2008, 4, 1),
        datetime(2008, 2, 29): datetime(2008, 4, 1),
        datetime(2008, 1, 15): datetime(2008, 4, 1),
        datetime(2008, 2, 27): datetime(2008, 4, 1),
        datetime(2008, 3, 15): datetime(2008, 4, 1),
        datetime(2007, 4, 1): datetime(2007, 4, 2),
        datetime(2007, 4, 2): datetime(2007, 4, 2),
        datetime(2007, 7, 1): datetime(2007, 7, 2),
        datetime(2007, 4, 15): datetime(2007, 7, 2),
        datetime(2007, 7, 2): datetime(2007, 7, 2)}))

    offset_cases.append((BQuarterBegin(startingMonth=1, n=-1), {
        datetime(2008, 1, 1): datetime(2007, 10, 1),
        datetime(2008, 1, 31): datetime(2008, 1, 1),
        datetime(2008, 2, 15): datetime(2008, 1, 1),
        datetime(2008, 2, 29): datetime(2008, 1, 1),
        datetime(2008, 3, 15): datetime(2008, 1, 1),
        datetime(2008, 3, 31): datetime(2008, 1, 1),
        datetime(2008, 4, 15): datetime(2008, 4, 1),
        datetime(2007, 7, 3): datetime(2007, 7, 2),
        datetime(2007, 4, 3): datetime(2007, 4, 2),
        datetime(2007, 7, 2): datetime(2007, 4, 2),
        datetime(2008, 4, 1): datetime(2008, 1, 1)}))

    offset_cases.append((BQuarterBegin(startingMonth=1, n=2), {
        datetime(2008, 1, 1): datetime(2008, 7, 1),
        datetime(2008, 1, 15): datetime(2008, 7, 1),
        datetime(2008, 2, 29): datetime(2008, 7, 1),
        datetime(2008, 3, 15): datetime(2008, 7, 1),
        datetime(2007, 3, 31): datetime(2007, 7, 2),
        datetime(2007, 4, 15): datetime(2007, 10, 1),
        datetime(2008, 4, 30): datetime(2008, 10, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)


class TestBQuarterEnd(Base):
    _offset = BQuarterEnd

    def test_repr(self):
        expected = "<BusinessQuarterEnd: startingMonth=3>"
        assert repr(BQuarterEnd()) == expected
        expected = "<BusinessQuarterEnd: startingMonth=3>"
        assert repr(BQuarterEnd(startingMonth=3)) == expected
        expected = "<BusinessQuarterEnd: startingMonth=1>"
        assert repr(BQuarterEnd(startingMonth=1)) == expected

    def test_isAnchored(self):
        assert BQuarterEnd(startingMonth=1).isAnchored()
        assert BQuarterEnd().isAnchored()
        assert not BQuarterEnd(2, startingMonth=1).isAnchored()

    def test_offset_corner_case(self):
        # corner
        offset = BQuarterEnd(n=-1, startingMonth=1)
        assert datetime(2010, 1, 31) + offset == datetime(2010, 1, 29)

    offset_cases = []
    offset_cases.append((BQuarterEnd(startingMonth=1), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 4, 30),
        datetime(2008, 2, 15): datetime(2008, 4, 30),
        datetime(2008, 2, 29): datetime(2008, 4, 30),
        datetime(2008, 3, 15): datetime(2008, 4, 30),
        datetime(2008, 3, 31): datetime(2008, 4, 30),
        datetime(2008, 4, 15): datetime(2008, 4, 30),
        datetime(2008, 4, 30): datetime(2008, 7, 31)}))

    offset_cases.append((BQuarterEnd(startingMonth=2), {
        datetime(2008, 1, 1): datetime(2008, 2, 29),
        datetime(2008, 1, 31): datetime(2008, 2, 29),
        datetime(2008, 2, 15): datetime(2008, 2, 29),
        datetime(2008, 2, 29): datetime(2008, 5, 30),
        datetime(2008, 3, 15): datetime(2008, 5, 30),
        datetime(2008, 3, 31): datetime(2008, 5, 30),
        datetime(2008, 4, 15): datetime(2008, 5, 30),
        datetime(2008, 4, 30): datetime(2008, 5, 30)}))

    offset_cases.append((BQuarterEnd(startingMonth=1, n=0), {
        datetime(2008, 1, 1): datetime(2008, 1, 31),
        datetime(2008, 1, 31): datetime(2008, 1, 31),
        datetime(2008, 2, 15): datetime(2008, 4, 30),
        datetime(2008, 2, 29): datetime(2008, 4, 30),
        datetime(2008, 3, 15): datetime(2008, 4, 30),
        datetime(2008, 3, 31): datetime(2008, 4, 30),
        datetime(2008, 4, 15): datetime(2008, 4, 30),
        datetime(2008, 4, 30): datetime(2008, 4, 30)}))

    offset_cases.append((BQuarterEnd(startingMonth=1, n=-1), {
        datetime(2008, 1, 1): datetime(2007, 10, 31),
        datetime(2008, 1, 31): datetime(2007, 10, 31),
        datetime(2008, 2, 15): datetime(2008, 1, 31),
        datetime(2008, 2, 29): datetime(2008, 1, 31),
        datetime(2008, 3, 15): datetime(2008, 1, 31),
        datetime(2008, 3, 31): datetime(2008, 1, 31),
        datetime(2008, 4, 15): datetime(2008, 1, 31),
        datetime(2008, 4, 30): datetime(2008, 1, 31)}))

    offset_cases.append((BQuarterEnd(startingMonth=1, n=2), {
        datetime(2008, 1, 31): datetime(2008, 7, 31),
        datetime(2008, 2, 15): datetime(2008, 7, 31),
        datetime(2008, 2, 29): datetime(2008, 7, 31),
        datetime(2008, 3, 15): datetime(2008, 7, 31),
        datetime(2008, 3, 31): datetime(2008, 7, 31),
        datetime(2008, 4, 15): datetime(2008, 7, 31),
        datetime(2008, 4, 30): datetime(2008, 10, 31)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [
        (BQuarterEnd(1, startingMonth=1), datetime(2008, 1, 31), True),
        (BQuarterEnd(1, startingMonth=1), datetime(2007, 12, 31), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2008, 2, 29), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2007, 3, 30), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2007, 3, 31), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2008, 4, 30), True),
        (BQuarterEnd(1, startingMonth=1), datetime(2008, 5, 30), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2007, 6, 29), False),
        (BQuarterEnd(1, startingMonth=1), datetime(2007, 6, 30), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2008, 1, 31), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2007, 12, 31), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2008, 2, 29), True),
        (BQuarterEnd(1, startingMonth=2), datetime(2007, 3, 30), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2007, 3, 31), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2008, 4, 30), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2008, 5, 30), True),
        (BQuarterEnd(1, startingMonth=2), datetime(2007, 6, 29), False),
        (BQuarterEnd(1, startingMonth=2), datetime(2007, 6, 30), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2008, 1, 31), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2007, 12, 31), True),
        (BQuarterEnd(1, startingMonth=3), datetime(2008, 2, 29), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2007, 3, 30), True),
        (BQuarterEnd(1, startingMonth=3), datetime(2007, 3, 31), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2008, 4, 30), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2008, 5, 30), False),
        (BQuarterEnd(1, startingMonth=3), datetime(2007, 6, 29), True),
        (BQuarterEnd(1, startingMonth=3), datetime(2007, 6, 30), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)

# --------------------------------------------------------------------
# Years


class TestYearBegin(Base):
    _offset = YearBegin

    def test_misspecified(self):
        pytest.raises(ValueError, YearBegin, month=13)

    offset_cases = []
    offset_cases.append((YearBegin(), {
        datetime(2008, 1, 1): datetime(2009, 1, 1),
        datetime(2008, 6, 30): datetime(2009, 1, 1),
        datetime(2008, 12, 31): datetime(2009, 1, 1),
        datetime(2005, 12, 30): datetime(2006, 1, 1),
        datetime(2005, 12, 31): datetime(2006, 1, 1)}))

    offset_cases.append((YearBegin(0), {
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2008, 6, 30): datetime(2009, 1, 1),
        datetime(2008, 12, 31): datetime(2009, 1, 1),
        datetime(2005, 12, 30): datetime(2006, 1, 1),
        datetime(2005, 12, 31): datetime(2006, 1, 1)}))

    offset_cases.append((YearBegin(3), {
        datetime(2008, 1, 1): datetime(2011, 1, 1),
        datetime(2008, 6, 30): datetime(2011, 1, 1),
        datetime(2008, 12, 31): datetime(2011, 1, 1),
        datetime(2005, 12, 30): datetime(2008, 1, 1),
        datetime(2005, 12, 31): datetime(2008, 1, 1)}))

    offset_cases.append((YearBegin(-1), {
        datetime(2007, 1, 1): datetime(2006, 1, 1),
        datetime(2007, 1, 15): datetime(2007, 1, 1),
        datetime(2008, 6, 30): datetime(2008, 1, 1),
        datetime(2008, 12, 31): datetime(2008, 1, 1),
        datetime(2006, 12, 29): datetime(2006, 1, 1),
        datetime(2006, 12, 30): datetime(2006, 1, 1),
        datetime(2007, 1, 1): datetime(2006, 1, 1)}))

    offset_cases.append((YearBegin(-2), {
        datetime(2007, 1, 1): datetime(2005, 1, 1),
        datetime(2008, 6, 30): datetime(2007, 1, 1),
        datetime(2008, 12, 31): datetime(2007, 1, 1)}))

    offset_cases.append((YearBegin(month=4), {
        datetime(2007, 4, 1): datetime(2008, 4, 1),
        datetime(2007, 4, 15): datetime(2008, 4, 1),
        datetime(2007, 3, 1): datetime(2007, 4, 1),
        datetime(2007, 12, 15): datetime(2008, 4, 1),
        datetime(2012, 1, 31): datetime(2012, 4, 1)}))

    offset_cases.append((YearBegin(0, month=4), {
        datetime(2007, 4, 1): datetime(2007, 4, 1),
        datetime(2007, 3, 1): datetime(2007, 4, 1),
        datetime(2007, 12, 15): datetime(2008, 4, 1),
        datetime(2012, 1, 31): datetime(2012, 4, 1)}))

    offset_cases.append((YearBegin(4, month=4), {
        datetime(2007, 4, 1): datetime(2011, 4, 1),
        datetime(2007, 4, 15): datetime(2011, 4, 1),
        datetime(2007, 3, 1): datetime(2010, 4, 1),
        datetime(2007, 12, 15): datetime(2011, 4, 1),
        datetime(2012, 1, 31): datetime(2015, 4, 1)}))

    offset_cases.append((YearBegin(-1, month=4), {
        datetime(2007, 4, 1): datetime(2006, 4, 1),
        datetime(2007, 3, 1): datetime(2006, 4, 1),
        datetime(2007, 12, 15): datetime(2007, 4, 1),
        datetime(2012, 1, 31): datetime(2011, 4, 1)}))

    offset_cases.append((YearBegin(-3, month=4), {
        datetime(2007, 4, 1): datetime(2004, 4, 1),
        datetime(2007, 3, 1): datetime(2004, 4, 1),
        datetime(2007, 12, 15): datetime(2005, 4, 1),
        datetime(2012, 1, 31): datetime(2009, 4, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(YearBegin(), datetime(2007, 1, 3), False),
                       (YearBegin(), datetime(2008, 1, 1), True),
                       (YearBegin(), datetime(2006, 12, 31), False),
                       (YearBegin(), datetime(2006, 1, 2), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestYearEnd(Base):
    _offset = YearEnd

    def test_misspecified(self):
        pytest.raises(ValueError, YearEnd, month=13)

    offset_cases = []
    offset_cases.append((YearEnd(), {
        datetime(2008, 1, 1): datetime(2008, 12, 31),
        datetime(2008, 6, 30): datetime(2008, 12, 31),
        datetime(2008, 12, 31): datetime(2009, 12, 31),
        datetime(2005, 12, 30): datetime(2005, 12, 31),
        datetime(2005, 12, 31): datetime(2006, 12, 31)}))

    offset_cases.append((YearEnd(0), {
        datetime(2008, 1, 1): datetime(2008, 12, 31),
        datetime(2008, 6, 30): datetime(2008, 12, 31),
        datetime(2008, 12, 31): datetime(2008, 12, 31),
        datetime(2005, 12, 30): datetime(2005, 12, 31)}))

    offset_cases.append((YearEnd(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 31),
        datetime(2008, 6, 30): datetime(2007, 12, 31),
        datetime(2008, 12, 31): datetime(2007, 12, 31),
        datetime(2006, 12, 29): datetime(2005, 12, 31),
        datetime(2006, 12, 30): datetime(2005, 12, 31),
        datetime(2007, 1, 1): datetime(2006, 12, 31)}))

    offset_cases.append((YearEnd(-2), {
        datetime(2007, 1, 1): datetime(2005, 12, 31),
        datetime(2008, 6, 30): datetime(2006, 12, 31),
        datetime(2008, 12, 31): datetime(2006, 12, 31)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(YearEnd(), datetime(2007, 12, 31), True),
                       (YearEnd(), datetime(2008, 1, 1), False),
                       (YearEnd(), datetime(2006, 12, 31), True),
                       (YearEnd(), datetime(2006, 12, 29), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestYearEndDiffMonth(Base):
    offset_cases = []
    offset_cases.append((YearEnd(month=3),
                        {datetime(2008, 1, 1): datetime(2008, 3, 31),
                         datetime(2008, 2, 15): datetime(2008, 3, 31),
                         datetime(2008, 3, 31): datetime(2009, 3, 31),
                         datetime(2008, 3, 30): datetime(2008, 3, 31),
                         datetime(2005, 3, 31): datetime(2006, 3, 31),
                         datetime(2006, 7, 30): datetime(2007, 3, 31)}))

    offset_cases.append((YearEnd(0, month=3),
                        {datetime(2008, 1, 1): datetime(2008, 3, 31),
                         datetime(2008, 2, 28): datetime(2008, 3, 31),
                         datetime(2008, 3, 31): datetime(2008, 3, 31),
                         datetime(2005, 3, 30): datetime(2005, 3, 31)}))

    offset_cases.append((YearEnd(-1, month=3),
                        {datetime(2007, 1, 1): datetime(2006, 3, 31),
                         datetime(2008, 2, 28): datetime(2007, 3, 31),
                         datetime(2008, 3, 31): datetime(2007, 3, 31),
                         datetime(2006, 3, 29): datetime(2005, 3, 31),
                         datetime(2006, 3, 30): datetime(2005, 3, 31),
                         datetime(2007, 3, 1): datetime(2006, 3, 31)}))

    offset_cases.append((YearEnd(-2, month=3),
                        {datetime(2007, 1, 1): datetime(2005, 3, 31),
                         datetime(2008, 6, 30): datetime(2007, 3, 31),
                         datetime(2008, 3, 31): datetime(2006, 3, 31)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(YearEnd(month=3), datetime(2007, 3, 31), True),
                       (YearEnd(month=3), datetime(2008, 1, 1), False),
                       (YearEnd(month=3), datetime(2006, 3, 31), True),
                       (YearEnd(month=3), datetime(2006, 3, 29), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestBYearBegin(Base):
    _offset = BYearBegin

    def test_misspecified(self):
        pytest.raises(ValueError, BYearBegin, month=13)
        pytest.raises(ValueError, BYearEnd, month=13)

    offset_cases = []
    offset_cases.append((BYearBegin(), {
        datetime(2008, 1, 1): datetime(2009, 1, 1),
        datetime(2008, 6, 30): datetime(2009, 1, 1),
        datetime(2008, 12, 31): datetime(2009, 1, 1),
        datetime(2011, 1, 1): datetime(2011, 1, 3),
        datetime(2011, 1, 3): datetime(2012, 1, 2),
        datetime(2005, 12, 30): datetime(2006, 1, 2),
        datetime(2005, 12, 31): datetime(2006, 1, 2)}))

    offset_cases.append((BYearBegin(0), {
        datetime(2008, 1, 1): datetime(2008, 1, 1),
        datetime(2008, 6, 30): datetime(2009, 1, 1),
        datetime(2008, 12, 31): datetime(2009, 1, 1),
        datetime(2005, 12, 30): datetime(2006, 1, 2),
        datetime(2005, 12, 31): datetime(2006, 1, 2)}))

    offset_cases.append((BYearBegin(-1), {
        datetime(2007, 1, 1): datetime(2006, 1, 2),
        datetime(2009, 1, 4): datetime(2009, 1, 1),
        datetime(2009, 1, 1): datetime(2008, 1, 1),
        datetime(2008, 6, 30): datetime(2008, 1, 1),
        datetime(2008, 12, 31): datetime(2008, 1, 1),
        datetime(2006, 12, 29): datetime(2006, 1, 2),
        datetime(2006, 12, 30): datetime(2006, 1, 2),
        datetime(2006, 1, 1): datetime(2005, 1, 3)}))

    offset_cases.append((BYearBegin(-2), {
        datetime(2007, 1, 1): datetime(2005, 1, 3),
        datetime(2007, 6, 30): datetime(2006, 1, 2),
        datetime(2008, 12, 31): datetime(2007, 1, 1)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)


class TestBYearEnd(Base):
    _offset = BYearEnd

    offset_cases = []
    offset_cases.append((BYearEnd(), {
        datetime(2008, 1, 1): datetime(2008, 12, 31),
        datetime(2008, 6, 30): datetime(2008, 12, 31),
        datetime(2008, 12, 31): datetime(2009, 12, 31),
        datetime(2005, 12, 30): datetime(2006, 12, 29),
        datetime(2005, 12, 31): datetime(2006, 12, 29)}))

    offset_cases.append((BYearEnd(0), {
        datetime(2008, 1, 1): datetime(2008, 12, 31),
        datetime(2008, 6, 30): datetime(2008, 12, 31),
        datetime(2008, 12, 31): datetime(2008, 12, 31),
        datetime(2005, 12, 31): datetime(2006, 12, 29)}))

    offset_cases.append((BYearEnd(-1), {
        datetime(2007, 1, 1): datetime(2006, 12, 29),
        datetime(2008, 6, 30): datetime(2007, 12, 31),
        datetime(2008, 12, 31): datetime(2007, 12, 31),
        datetime(2006, 12, 29): datetime(2005, 12, 30),
        datetime(2006, 12, 30): datetime(2006, 12, 29),
        datetime(2007, 1, 1): datetime(2006, 12, 29)}))

    offset_cases.append((BYearEnd(-2), {
        datetime(2007, 1, 1): datetime(2005, 12, 30),
        datetime(2008, 6, 30): datetime(2006, 12, 29),
        datetime(2008, 12, 31): datetime(2006, 12, 29)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    on_offset_cases = [(BYearEnd(), datetime(2007, 12, 31), True),
                       (BYearEnd(), datetime(2008, 1, 1), False),
                       (BYearEnd(), datetime(2006, 12, 31), False),
                       (BYearEnd(), datetime(2006, 12, 29), True)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)


class TestBYearEndLagged(Base):
    _offset = BYearEnd

    def test_bad_month_fail(self):
        pytest.raises(Exception, BYearEnd, month=13)
        pytest.raises(Exception, BYearEnd, month=0)

    offset_cases = []
    offset_cases.append((BYearEnd(month=6), {
        datetime(2008, 1, 1): datetime(2008, 6, 30),
        datetime(2007, 6, 30): datetime(2008, 6, 30)}))

    offset_cases.append((BYearEnd(n=-1, month=6), {
        datetime(2008, 1, 1): datetime(2007, 6, 29),
        datetime(2007, 6, 30): datetime(2007, 6, 29)}))

    @pytest.mark.parametrize('case', offset_cases)
    def test_offset(self, case):
        offset, cases = case
        for base, expected in compat.iteritems(cases):
            assert_offset_equal(offset, base, expected)

    def test_roll(self):
        offset = BYearEnd(month=6)
        date = datetime(2009, 11, 30)

        assert offset.rollforward(date) == datetime(2010, 6, 30)
        assert offset.rollback(date) == datetime(2009, 6, 30)

    on_offset_cases = [(BYearEnd(month=2), datetime(2007, 2, 28), True),
                       (BYearEnd(month=6), datetime(2007, 6, 30), False)]

    @pytest.mark.parametrize('case', on_offset_cases)
    def test_onOffset(self, case):
        offset, dt, expected = case
        assert_onOffset(offset, dt, expected)



# coding: utf-8

from pandas.io.msgpack import ExtType, packb, unpackb


def test_str8():
    header = b'\xd9'
    data = b'x' * 32
    b = packb(data.decode(), use_bin_type=True)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\x20'
    assert b[2:] == data
    assert unpackb(b) == data

    data = b'x' * 255
    b = packb(data.decode(), use_bin_type=True)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\xff'
    assert b[2:] == data
    assert unpackb(b) == data


def test_bin8():
    header = b'\xc4'
    data = b''
    b = packb(data, use_bin_type=True)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\x00'
    assert b[2:] == data
    assert unpackb(b) == data

    data = b'x' * 255
    b = packb(data, use_bin_type=True)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\xff'
    assert b[2:] == data
    assert unpackb(b) == data


def test_bin16():
    header = b'\xc5'
    data = b'x' * 256
    b = packb(data, use_bin_type=True)
    assert len(b) == len(data) + 3
    assert b[0:1] == header
    assert b[1:3] == b'\x01\x00'
    assert b[3:] == data
    assert unpackb(b) == data

    data = b'x' * 65535
    b = packb(data, use_bin_type=True)
    assert len(b) == len(data) + 3
    assert b[0:1] == header
    assert b[1:3] == b'\xff\xff'
    assert b[3:] == data
    assert unpackb(b) == data


def test_bin32():
    header = b'\xc6'
    data = b'x' * 65536
    b = packb(data, use_bin_type=True)
    assert len(b) == len(data) + 5
    assert b[0:1] == header
    assert b[1:5] == b'\x00\x01\x00\x00'
    assert b[5:] == data
    assert unpackb(b) == data


def test_ext():
    def check(ext, packed):
        assert packb(ext) == packed
        assert unpackb(packed) == ext

    check(ExtType(0x42, b'Z'), b'\xd4\x42Z')  # fixext 1
    check(ExtType(0x42, b'ZZ'), b'\xd5\x42ZZ')  # fixext 2
    check(ExtType(0x42, b'Z' * 4), b'\xd6\x42' + b'Z' * 4)  # fixext 4
    check(ExtType(0x42, b'Z' * 8), b'\xd7\x42' + b'Z' * 8)  # fixext 8
    check(ExtType(0x42, b'Z' * 16), b'\xd8\x42' + b'Z' * 16)  # fixext 16
    # ext 8
    check(ExtType(0x42, b''), b'\xc7\x00\x42')
    check(ExtType(0x42, b'Z' * 255), b'\xc7\xff\x42' + b'Z' * 255)
    # ext 16
    check(ExtType(0x42, b'Z' * 256), b'\xc8\x01\x00\x42' + b'Z' * 256)
    check(ExtType(0x42, b'Z' * 0xffff), b'\xc8\xff\xff\x42' + b'Z' * 0xffff)
    # ext 32
    check(
        ExtType(0x42, b'Z' *
                0x10000), b'\xc9\x00\x01\x00\x00\x42' + b'Z' * 0x10000)
    # needs large memory
    # check(ExtType(0x42, b'Z'*0xffffffff),
    #              b'\xc9\xff\xff\xff\xff\x42' + b'Z'*0xffffffff)



# -*- coding: utf-8 -*-
"""Tests that the tslibs API is locked down"""

from pandas._libs import tslibs


def test_namespace():

    submodules = ['ccalendar',
                  'conversion',
                  'fields',
                  'frequencies',
                  'nattype',
                  'np_datetime',
                  'offsets',
                  'parsing',
                  'period',
                  'resolution',
                  'strptime',
                  'timedeltas',
                  'timestamps',
                  'timezones']

    api = ['NaT',
           'iNaT',
           'is_null_datetimelike',
           'OutOfBoundsDatetime',
           'Period',
           'IncompatibleFrequency',
           'Timedelta',
           'Timestamp',
           'delta_to_nanoseconds',
           'ints_to_pytimedelta',
           'localize_pydatetime',
           'normalize_date',
           'tz_convert_single']

    expected = set(submodules + api)
    names = [x for x in dir(tslibs) if not x.startswith('__')]
    assert set(names) == expected
